"""thermo_models contains tools for running MDF and displaying results."""
# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from typing import Optional

import numpy as np
from equilibrator_api import Q_, ureg
from matplotlib import pyplot as plt
from matplotlib.collections import LineCollection

from .analysis_solution import PathwayAnalysisSolution


class PathwayMdfSolution(PathwayAnalysisSolution):
    """Handle MDF results.

    PathwayMDFData is a container class for MDF results, with plotting
    capabilities.
    """

    def __init__(
        self,
        thermo_model: "PathwayThermoModel",
        score: float,
        ln_conc: np.array,
        y: np.array,
        reaction_prices: np.array,
        compound_prices: np.array,
    ) -> None:
        """

        Parameters
        ----------
        thermo_model : ThermodynamicModel
        score : float
            the optimal Max-min Driving Force (in kJ/mol)
        ln_conc : ndarray
            log concentrations at MDF optimum
        y : ndarray
            uncertainty matrix coefficients at MDF optimum
        reaction_prices : ndarray
            shadow prices of reactions
        compound_prices : ndarray
            shadow prices of compound concentrations
        """
        super(PathwayMdfSolution, self).__init__(
            thermo_model, score, ln_conc, y
        )
        self.reaction_df["shadow_price"] = reaction_prices
        self.compound_df["shadow_price"] = compound_prices

    def plot_concentrations(self, ax: Optional[plt.Axes] = None) -> None:
        """Plot compound concentrations.

        :return: matplotlib Figure
        """
        ureg.setup_matplotlib(True)

        if not ax:
            fig, ax = plt.subplots(
                1, 1, figsize=(10, self.compound_df.shape[0] * 0.25)
            )

        data_df = self.compound_df.copy()
        data_df["y"] = np.arange(0, data_df.shape[0])
        data_df["shadow_sign"] = data_df.shadow_price.apply(np.sign)

        for shadow_sign, group_df in data_df.groupby("shadow_sign"):
            if shadow_sign == -1:
                color, marker, label = ("blue", "<", "shadow price < 0")
            elif shadow_sign == 1:
                color, marker, label = ("red", ">", "shadow price > 0")
            else:
                color, marker, label = ("grey", "d", "shadow price = 0")

            group_df.plot.scatter(
                x="concentration",
                y="y",
                s=40,
                c=color,
                marker=marker,
                ax=ax,
                zorder=2,
                colorbar=False,
                label=label,
            )
        ax.set_ylabel("")
        ax.set_yticks(data_df.y)

        for j, row in enumerate(data_df.itertuples(index=True)):
            ax.plot(
                [row.lower_bound.m_as("M"), row.upper_bound.m_as("M")],
                [row.y, row.y],
                color="lightgrey",
                linewidth=3,
                zorder=1,
                label="allowed range" if j == 0 else None,
            )

        ax.set_yticklabels(data_df.compound_id, fontsize=9)
        ax.set_xlabel("Concentration (M)")
        ax.set_xscale("log")
        ax.set_ylim(-0.5, data_df.shape[0] + 0.5)
        ax.legend(bbox_to_anchor=(1.02, 1), loc="upper left")

    def plot_driving_forces(self, ax: Optional[plt.Axes] = None) -> None:
        """Plot cumulative delta-G profiles.

        :return: matplotlib Figure
        """
        ureg.setup_matplotlib(True)

        if not ax:
            fig, ax = plt.subplots(
                1, 1, figsize=(3 + self.reaction_df.shape[0] * 0.3, 6)
            )

        data_df = self.reaction_df.copy()
        data_df.reindex()

        data_df[
            "cml_dgm"
        ] = self.reaction_df.physiological_dg_prime.cumsum().apply(
            lambda x: x.m_as("kJ/mol")
        )
        data_df[
            "cml_dg_opt"
        ] = self.reaction_df.optimized_dg_prime.cumsum().apply(
            lambda x: x.m_as("kJ/mol")
        )

        xticks = 0.5 + np.arange(data_df.shape[0])
        xticklabels = data_df.reaction_id.tolist()

        yvalues_phy = [0.0] + data_df.cml_dgm.tolist()
        yvalues_opt = [0.0] + data_df.cml_dg_opt.tolist()

        ax.plot(
            yvalues_phy,
            label="Physiological concentrations (1 mM)",
            color="lightgreen",
            linestyle="--",
            zorder=1,
        )
        ax.plot(
            yvalues_opt,
            label="MDF-optimized concentrations",
            color="grey",
            linewidth=2,
            zorder=2,
        )
        lines = [
            ((i, yvalues_opt[i]), (i + 1, yvalues_opt[i + 1]))
            for i in data_df[data_df.shadow_price != 0].index
        ]
        lines = LineCollection(
            lines,
            label="Bottleneck reactions",
            linewidth=3,
            color="red",
            linestyle="-",
            zorder=3,
            alpha=1,
        )
        ax.add_collection(lines)
        ax.set_xticks(xticks)
        ax.set_xticklabels(xticklabels, rotation=45, ha="center")
        ax.set_xlim(0, data_df.shape[0])

        ax.set_xlabel("Reaction Step")
        ax.set_ylabel(r"Cumulative $\Delta_r G^\prime$ (kJ/mol)")
        ax.legend(loc="best")
        ax.set_title(f"MDF = {self.score:.2f} kJ/mol")



"""analyze pathways using thermodynamic models."""
# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from typing import Callable, Dict, List, Optional, Union

import numpy as np
import pandas as pd
from equilibrator_api import Q_, ComponentContribution, Reaction
from equilibrator_cache import Compound
from sbtab import SBtab

from . import Bounds, StoichiometricModel


class Pathway(StoichiometricModel):
    """A pathway parsed from user input.

    Designed for checking input prior to converting to a stoichiometric model.
    """

    def __init__(
        self,
        reactions: List[Reaction],
        fluxes: Q_,
        comp_contrib: Optional[ComponentContribution] = None,
        standard_dg_primes: Optional[Q_] = None,
        dg_sigma: Optional[Q_] = None,
        bounds: Optional[Bounds] = None,
        config_dict: Optional[Dict[str, str]] = None,
        compound_id_mapping: Callable[[Compound], str] = None,
    ) -> None:
        """Initialize a Pathway object.

        Parameters
        ----------
        reactions : List[Reaction]
            a list of Reaction objects
        fluxes : Quantity
            relative fluxes in same order as
        comp_contrib : ComponentContribution
            a ComponentContribution object
        standard_dg_primes : Quantity, optional
            reaction energies (in kJ/mol)
        dg_sigma : Quantity, optional
            square root of the uncertainty covariance matrix (in kJ/mol)
        bounds : Bounds, optional
            bounds on metabolite concentrations (by default uses the
            "data/cofactors.csv" file in `equilibrator-api`)
        config_dict : dict, optional
            configuration parameters for Pathway analysis
        compound_id_mapping : callable, optional
            a function mapping compounds to their names in the model
        """
        super(Pathway, self).__init__(
            reactions=reactions,
            comp_contrib=comp_contrib,
            standard_dg_primes=standard_dg_primes,
            dg_sigma=dg_sigma,
            bounds=bounds,
            config_dict=config_dict,
            compound_id_mapping=compound_id_mapping,
        )

        assert fluxes.unitless or fluxes.check("[concentration]/[time]")
        self.fluxes = fluxes.flatten()
        assert self.fluxes.shape == (self.Nr,)
        self.I_dir = np.diag(np.sign(self.fluxes.magnitude).flat)

    @property
    def net_reaction(self) -> Reaction:
        """Calculate the sum of all the reactions in the pathway.

        :return: the net reaction
        """
        net_rxn_stoich = self.S @ self.fluxes.magnitude
        net_rxn_stoich = net_rxn_stoich[net_rxn_stoich != 0]
        sparse = net_rxn_stoich.to_dict()
        return Reaction(sparse)

    @property
    def net_reaction_formula(self) -> str:
        """Calculate the sum of all the reactions in the pathway.

        :return: the net reaction formula
        """
        return self._reaction_to_formula(self.net_reaction)

    @classmethod
    def from_network_sbtab(
        cls,
        filename: Union[str, SBtab.SBtabDocument],
        comp_contrib: Optional[ComponentContribution] = None,
        freetext: bool = True,
        bounds: Optional[Bounds] = None,
    ) -> object:
        """Initialize a Pathway object using a 'network'-only SBtab.

        Parameters
        ----------
        filename : str, SBtabDocument
            a filename containing an SBtabDocument (or the SBtabDocument
            object itself) defining the network (topology) only
        comp_contrib : ComponentContribution, optional
            a ComponentContribution object needed for parsing and searching
            the reactions. also used to set the aqueous parameters (pH, I, etc.)
        freetext : bool, optional
            a flag indicating whether the reactions are given as free-text (i.e.
            common names for compounds) or by standard database accessions
            (Default value: `True`)
        bounds : Bounds, optional
            bounds on metabolite concentrations (by default uses the
            "data/cofactors.csv" file in `equilibrator-api`)

        Returns
        -------
            a Pathway object
        """
        (
            reaction_df,
            comp_contrib,
            config_dict,
        ) = StoichiometricModel._read_network_sbtab(
            filename, comp_contrib, freetext
        )
        fluxes = reaction_df.RelativeFlux.apply(float).values * Q_(
            "dimensionless"
        )
        pp = Pathway(
            reactions=reaction_df.Reaction.tolist(),
            fluxes=fluxes,
            comp_contrib=comp_contrib,
            bounds=bounds,
            config_dict=config_dict,
        )
        return pp

    @classmethod
    def from_stoichiometric_model(
        cls, stoich_model: StoichiometricModel
    ) -> "Pathway":
        """Convert a StoichiometricModel into a Pathway.

        Assume all fluxes are 1.
        """
        fluxes = np.ones(len(stoich_model.reactions)) * Q_(1)
        pp = Pathway(
            reactions=stoich_model.reactions,
            fluxes=fluxes,
            comp_contrib=stoich_model.comp_contrib,
            standard_dg_primes=stoich_model.standard_dg_primes,
            dg_sigma=stoich_model.dg_sigma,
            bounds=stoich_model._bounds,
            config_dict=stoich_model.config_dict,
        )
        return pp

    @classmethod
    def from_sbtab(
        cls,
        filename: Union[str, SBtab.SBtabDocument],
        comp_contrib: Optional[ComponentContribution] = None,
    ) -> "Pathway":
        """Parse and SBtabDocument and return a StoichiometricModel.

        Parameters
        ----------
        filename : str or SBtabDocument
            a filename containing an SBtabDocument (or the SBtabDocument
            object itself) defining the pathway
        comp_contrib : ComponentContribution, optional
            a ComponentContribution object needed for parsing and searching
            the reactions. also used to set the aqueous parameters (pH, I, etc.)

        Returns
        -------
        pathway: Pathway
            A Pathway object based on the configuration SBtab

        """
        comp_contrib = comp_contrib or ComponentContribution()

        (
            sbtabdoc,
            reactions,
            reaction_ids,
            standard_dg_primes,
            dg_sigma,
            bounds,
            config_dict,
            compound_to_id,
        ) = cls._read_model_sbtab(filename, comp_contrib)

        # Read the Flux table
        # ---------------------------
        flux_sbtab = sbtabdoc.get_sbtab_by_id("Flux")
        assert flux_sbtab is not None, "Cannot find a 'Flux' table in the SBtab"
        flux_df = flux_sbtab.to_data_frame()
        assert set(flux_df.Reaction.unique()) == set(reaction_ids), (
            r"'Reaction' and 'Flux' tables are not 100% compatible (check"
            r"that the list of IDs is the same"
        )

        fluxes = (
            flux_df.set_index("Reaction")
            .loc[reaction_ids, "Value"]
            .apply(float)
        )

        fluxes = np.array(fluxes.values, ndmin=2, dtype=float).T

        try:
            # convert fluxes to M/s if they are in some other absolute unit
            flux_unit = flux_sbtab.get_attribute("Unit")
            fluxes *= Q_(1.0, flux_unit)
        except SBtab.SBtabError:
            # otherwise, assume these are relative fluxes
            fluxes *= Q_("dimensionless")

        return cls(
            reactions=reactions,
            fluxes=fluxes,
            comp_contrib=comp_contrib,
            standard_dg_primes=standard_dg_primes,
            dg_sigma=None,
            bounds=bounds,
            config_dict=config_dict,
            compound_id_mapping=compound_to_id.get,
        )

    def to_sbtab(self) -> SBtab.SBtabDocument:
        """Export the pathway to an SBtabDocument."""
        sbtabdoc = super(Pathway, self).to_sbtab()

        # add the flux table
        flux_df = pd.DataFrame(
            data=[
                ("rate of reaction", rxn.rid, f"{flux:.3g}")
                for rxn, flux in zip(self.reactions, self.fluxes.magnitude)
            ],
            columns=["QuantityType", "Reaction", "Value"],
        )
        flux_sbtab = SBtab.SBtabTable.from_data_frame(
            df=flux_df, table_id="Flux", table_type="Quantity"
        )
        flux_sbtab.change_attribute("Unit", self.fluxes.units)
        sbtabdoc.add_sbtab(flux_sbtab)

        return sbtabdoc



# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import itertools
import logging
from typing import Iterable, Optional, Tuple

import cvxpy as cp
import numpy as np
import pandas as pd
from cvxpy.constraints.constraint import Constraint
from cvxpy.expressions.expression import Expression
from cvxpy.problems.objective import Objective
from equilibrator_api import Q_, R, default_T

from .util import ECF_DEFAULTS


class EnzymeCostFunction(object):

    ECF_LEVEL_NAMES = [
        "capacity [M]",
        "thermodynamic",
        "saturation",
        "allosteric",
    ]
    CONFIDENCE_INTERVAL = 1.96  # TODO: move to params dict
    MINIMAL_STDEV = 1e-3  # for metabolites to be considered "fixed"
    DRIVING_FORCE_LB = 1e-3  # in kJ/mol

    def __init__(
        self,
        S: np.ndarray,
        fluxes: Q_,
        kcat: Q_,
        standard_dg: Q_,
        KMM: Q_,
        ln_conc_lb: np.ndarray,
        ln_conc_ub: np.ndarray,
        mw_enz: Optional[Q_] = None,
        mw_met: Optional[Q_] = None,
        A_act: Optional[np.ndarray] = None,
        A_inh: Optional[np.ndarray] = None,
        K_act: Optional[Q_] = None,
        K_inh: Optional[Q_] = None,
        idx_water: int = -1,
        params: Optional[dict] = None,
    ):
        """Create a Cost Function object.

        Parameters
        ----------
        S: ndarray
            stoichiometric matrix [unitless]
        fluxes: Quantity [concentration]/[time]
            steady-state fluxes [flux units]
        kcat: Quantity [1/time]
            turnover numbers
        standard_dg: Quantity [energy]/[substance]
            standard Gibbs free energies of reaction
        KMM: Quantity [concentration]
            Michaelis-Menten coefficients
        ln_conc_lb: ndarray
            lower bounds on metabolite concentrations [ln M]
        ln_conc_ub: ndarray
            upper bounds on metabolite concentrations [ln M]
        mw_enz: Quantity, optional [mass]
            enzyme molecular weights
        mw_met: Quantity, optional [mass]
            metabolite molecular weights
        A_act: ndarray, optional
            Hill coefficient matrix of allosteric activators
        A_inh: ndarray, optional
            Hill coefficient matrix of allosteric inhibitors
        K_act: Quantity, optional [concentration]
            affinity coefficient matrix of allosteric activators
        K_inh: Quantity, optional [concentration]
            affinity coefficient matrix of allosteric inhibitors
        idx_water: int
            the index of water in the stoichiometric matrix (or -1
            if water is not part of the model)
        params: dict, optional
            dictionary of extra parameters
        """
        self.params = dict(ECF_DEFAULTS)
        if params is not None:
            self.params.update(params)

        self.S = S
        self.idx_water = idx_water

        if fluxes.check("[concentration]/[time]"):
            self.fluxes = fluxes.m_as("M/s").flatten()
        elif fluxes.unitless:
            # relative fluxes are dimensionless
            self.fluxes = fluxes.m_as("").flatten()
        else:
            raise ValueError("Fluxes must be in units of M/s or dimensionless")
        assert (self.fluxes > 0.0).all()

        assert kcat.check("1/[time]")
        self.kcat = kcat.m_as("1/s").flatten()

        assert standard_dg.check("[energy]/[substance]")
        self.standard_dg_over_rt = (
            (standard_dg / (R * default_T)).m_as("").flatten()
        )

        assert KMM.check("[concentration]")
        self.KMM = KMM.m_as("M")

        self.ln_conc_lb = ln_conc_lb.flatten()
        self.ln_conc_ub = ln_conc_ub.flatten()

        # In MDMC we use Z-score to describe distribution within the metabolite
        # so we need to convert the "hard" bounds into a Gaussian distribution
        # we assume that the given bounds represent the 95% confidence interval
        # a the Gaussian distribution (i.e. [mu - 1.96*sigma, mu + 1.96*sigma])
        # Therefore:
        #              mu = (ub + lb) / 2
        #           sigma = (ub - lb) / 3.92
        self.ln_conc_mu = (self.ln_conc_ub + self.ln_conc_lb) / 2.0
        self.ln_conc_sigma = (self.ln_conc_ub - self.ln_conc_lb) / (
            self.CONFIDENCE_INTERVAL * 2.0
        )

        self.Nc, self.Nr = S.shape
        assert self.fluxes.shape == (self.Nr,)
        assert self.kcat.shape == (self.Nr,)
        assert self.standard_dg_over_rt.shape == (self.Nr,)
        assert self.KMM.shape == (self.Nc, self.Nr)
        assert self.ln_conc_lb.shape == (self.Nc,)
        assert self.ln_conc_ub.shape == (self.Nc,)

        self.cids = ["C%04d" % i for i in range(self.Nc)]

        self.S_subs = abs(self.S)
        self.S_prod = abs(self.S)
        self.S_subs[self.S > 0] = 0
        self.S_prod[self.S < 0] = 0

        # if the kcat source is 'gmean' we need to recalculate the
        # kcat_fwd using the formula:
        # kcat_fwd = kcat_gmean * sqrt(kEQ * prod_S(KMM) / prod_P(KMM))

        if self.params["kcat_source"] == "gmean":
            ln_KMM_prod = np.array(np.diag(self.S.T @ np.log(self.KMM)))
            ln_ratio = -ln_KMM_prod - self.standard_dg_over_rt
            factor = np.sqrt(np.exp(ln_ratio))
            self.kcat *= factor

        # molecular weights of enzymes and metabolites
        if mw_enz is None:
            self.mw_enz = np.ones(self.Nr)
        else:
            assert mw_met.check("[mass]")
            self.mw_enz = mw_enz.m_as("Da").flatten()
            assert self.mw_enz.shape == (self.Nr,)
            assert (self.mw_enz > 0.0).all()

        if mw_met is None:
            self.mw_met = np.ones(self.Nc)
        else:
            assert mw_met.check("[mass]")
            self.mw_met = mw_met.m_as("Da").flatten()
            assert self.mw_met.shape == (self.Nc,)
            assert (self.mw_met > 0.0).all()

        # allosteric regulation term

        if A_act is None or K_act is None:
            self.A_act = np.zeros(S.shape)
            self.K_act = np.ones(S.shape)
        else:
            assert S.shape == A_act.shape
            assert S.shape == K_act.shape
            assert K_act.check("[concentration]")
            self.A_act = A_act
            self.K_act = K_act.m_as("M")

        if A_inh is None or K_inh is None:
            self.A_inh = np.zeros(S.shape)
            self.K_inh = np.ones(S.shape)
        else:
            assert S.shape == A_inh.shape
            assert S.shape == K_inh.shape
            assert K_inh.check("[concentration]")
            self.A_inh = A_inh
            self.K_inh = K_inh.m_as("M")

        # if one of the compounds is water, we remove its effect on the
        # saturation, and the MW cost of metabolites
        if self.idx_water > -1:
            self.S_subs[self.idx_water, :] = 0
            self.S_prod[self.idx_water, :] = 0

        # preprocessing: these auxiliary matrices help calculate the ECF3 and
        # ECF4 faster
        self.act_denom = np.diag(self.A_act.T @ np.log(self.K_act))
        self.inh_denom = np.diag(self.A_inh.T @ np.log(self.K_inh))

        try:
            self.ECF = eval("self._ECF%s" % self.params["version"])
        except AttributeError:
            raise ValueError(
                "The enzyme cost function %d is unknown"
                % self.params["version"]
            )

        assert (
            self.params.get("regularization", "volume").lower() == "volume"
        ), (
            "Support to regularizations other than 'volume' was dropped when"
            "moving to CVXPY"
        )
        self.ln_capacity = cp.Constant(np.log(self.fluxes / self.kcat))

    def _driving_forces(self, ln_conc: cp.Variable) -> cp.Expression:
        """
        calculate the driving force for every reaction in every condition
        """
        return -self.standard_dg_over_rt - self.S.T @ ln_conc

    def _ln_eta_thermodynamic(self, ln_conc: cp.Variable) -> cp.Expression:
        """Calculate the minus log value of the thermodynamic efficiencies."""
        return cp.log(1.0 - cp.exp(-self._driving_forces(ln_conc)))

    def _B_matrix(
        self, col_subs: np.ndarray, col_prod: np.ndarray
    ) -> np.ndarray:
        """Build the B matrix for the eta^kin expression.

        row_subs : np.ndarray
            A column from the substrate stoichiometric matrix. We assume
            coefficients represent reactant molecularities so
            only integer values are allowed.

        row_prod : np.ndarray
            A column from the product stoichiometric matrix. We assume
            coefficients represent reactant molecularities so
            only integer values are allowed.
        """

        def K_matrix(n: int) -> np.ndarray:
            """Make the 'K' matrix for the CM rate law."""
            lst = list(itertools.product([0, 1], repeat=n))
            lst.pop(0)  # remove the [0, 0, ..., 0] row
            return np.array(lst)

        def expand_S(coeffs: np.ndarray) -> np.ndarray:
            """Expand a coefficient column into a matrix with duplicates."""
            cs = list(np.cumsum(list(map(int, coeffs.flat))))
            S_tmp = np.zeros((cs[-1], self.Nc))
            for j, (i_from, i_to) in enumerate(zip([0] + cs, cs)):
                S_tmp[i_from:i_to, j] = 1
            return S_tmp

        S_subs = expand_S(col_subs)
        S_prod = expand_S(col_prod)

        A = np.vstack(
            [
                np.zeros((1, self.Nc)),
                K_matrix(S_subs.shape[0]) @ S_subs,
                K_matrix(S_prod.shape[0]) @ S_prod,
            ]
        )

        return A - np.ones((A.shape[0], S_subs.shape[0])) @ S_subs

    def _ln_D_CM_minus_ln_D_S(self, i: int, ln_conc: cp.Variable) -> Expression:
        B = self._B_matrix(self.S_subs[:, i], self.S_prod[:, i])
        return cp.log_sum_exp(B @ (ln_conc - np.log(self.KMM[:, i])))

    def _ln_eta_kinetic(self, ln_conc: cp.Variable) -> Expression:
        """
        the kinetic part of ECF3 and ECF4
        """
        ln_KMM = np.log(self.KMM)
        ln_conc_mat = cp.vstack([ln_conc] * self.Nr).T

        ln_D_S = self.S_subs.T @ (ln_conc_mat - ln_KMM)
        ln_D_P = self.S_prod.T @ (ln_conc_mat - ln_KMM)

        if self.params["denominator"] == "S":
            return cp.Constant(np.zeros(self.Nr))
        elif self.params["denominator"] == "1S":
            return -cp.diag(cp.logistic(-ln_D_S))
        elif self.params["denominator"] == "SP":
            return -cp.diag(cp.logistic(-ln_D_S + ln_D_P))
        elif self.params["denominator"] == "1SP":
            return -cp.diag(cp.logistic(-ln_D_S + cp.logistic(ln_D_P)))
        elif self.params["denominator"] == "CM":
            ln_eta = [
                -self._ln_D_CM_minus_ln_D_S(i, ln_conc) for i in range(self.Nr)
            ]
            return cp.reshape(cp.vstack(ln_eta), (self.Nr,))
        else:
            raise ValueError(
                "unsupported denominator: " + self.params["denominator"]
            )

    def _ln_eta_allosteric(self, ln_conc: cp.Variable) -> Expression:
        ln_K_act = np.log(self.K_act)
        ln_K_inh = np.log(self.K_inh)
        ln_conc_mat = cp.vstack([ln_conc] * self.Nr).T

        # original allosteric term:
        # 1 / (1 + prod(k_a/c_a)) / (1 + prod(c_i/k_i))
        # ln_act = self.A_act.T @ (ln_conc_mat - ln_K_act)
        # ln_inh = self.A_inh.T @ (ln_conc_mat - ln_K_inh)
        # ln_eta_act = -cp.diag(cp.logistic(-ln_act))
        # ln_eta_inh = -cp.diag(cp.logistic(ln_inh))

        # new allosteric term:
        # 1 / prod(1 + k_a/c_a) / prod(1 + c_i/k_i)
        ln_act = self.A_act.T @ cp.logistic(ln_K_act - ln_conc_mat)
        ln_inh = self.A_inh.T @ cp.logistic(ln_conc_mat - ln_K_inh)
        ln_eta_act = -cp.diag(ln_act)
        ln_eta_inh = -cp.diag(ln_inh)

        return ln_eta_act + ln_eta_inh

    def is_feasible(self, ln_conc: np.ndarray) -> bool:
        df = self._driving_forces(ln_conc)
        return (df > 0).all()

    def get_fluxes(self, ln_conc: np.ndarray, E: np.ndarray) -> np.ndarray:
        """Calculate the fluxes based on metabolite and enzyme conc."""
        assert ln_conc.shape == (self.Nc,)
        assert E.shape == (self.Nr,)

        v = self.kcat * E
        v *= np.exp(self._ln_eta_thermodynamic(ln_conc).value)
        v *= np.exp(self._ln_eta_kinetic(ln_conc).value)
        v *= np.exp(self._ln_eta_allosteric(ln_conc))

        return v.squeeze()

    def _ECF1(self, ln_conc: cp.Variable) -> Expression:
        """
        Arguments:
            A single metabolite ln-concentration vector

        Returns:
            The most basic Enzyme Cost Function (only dependent on flux
            and kcat). Gives the predicted enzyme concentrations in [M]
        """
        # ln_conc is not used for ECF1
        return self.ln_capacity

    def _ECF2(self, ln_conc: cp.Variable) -> Expression:
        """
        Arguments:
            A single metabolite ln-concentration vector

        Returns:
            The thermodynamic-only Enzyme Cost Function.
            Gives the predicted enzyme concentrations in [M].
        """
        return self._ECF1(ln_conc) - self._ln_eta_thermodynamic(ln_conc)

    def _ECF3(self, ln_conc: cp.Variable) -> Expression:
        """
        Arguments:
            A single metabolite ln-concentration vector

        Returns:
            An Enzyme Cost Function that integrates kinetic and
            thermodynamic data, but no allosteric regulation.
            Gives the predicted enzyme concentrations in [M].
        """
        return self._ECF2(ln_conc) - self._ln_eta_kinetic(ln_conc)

    def _ECF4(self, ln_conc: cp.Variable) -> Expression:
        """
        Arguments:
            A single metabolite ln-concentration vector

        Returns:
            The full Enzyme Cost Function, i.e. with kinetic, thermodynamic
            and allosteric data.
            Gives the predicted enzyme concentrations in [M].
        """
        return self._ECF3(ln_conc) - self._ln_eta_allosteric(ln_conc)

    def get_enzyme_cost_partitions(self, ln_conc: np.ndarray) -> np.ndarray:
        """
        Arguments:
            A single metabolite ln-concentration vector

        Returns:
            A matrix contining the enzyme costs separated to the 4 ECF
            factors (as columns).
            The first column is the ECF1 predicted concentrations in [M].
            The other columns are unitless (added cost, always > 1)
        """
        cap = cp.exp(self.ln_capacity)  # capacity
        trm = cp.exp(-self._ln_eta_thermodynamic(ln_conc))  # thermodynamics
        kin = cp.exp(-self._ln_eta_kinetic(ln_conc))  # kinetics
        alo = cp.exp(-self._ln_eta_allosteric(ln_conc))

        return cp.vstack([cap, trm, kin, alo]).T.value

    def get_volumes(self, ln_conc: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Arguments:
            A single metabolite ln-concentration vector

        Returns:
            Two arrays containing the enzyme volumes and
            metabolite volumes (at the provided point)
        """
        enz_conc = np.exp(self.ECF(ln_conc).value)
        met_conc = np.exp(ln_conc)
        enz_vols = np.multiply(enz_conc, self.mw_enz)
        met_vols = np.multiply(met_conc, self.mw_met)

        # we have to remove the volume of water, otherwise it dominates all
        # other volumes
        met_vols[self.idx_water] = 0.0
        return enz_vols, met_vols

    def _ln_total_enzyme_weight(self, ln_conc: cp.Variable) -> Expression:
        """Calculate the enzyme cost for an input concentration profile."""
        return cp.log_sum_exp(self.ECF(ln_conc) + np.log(self.mw_enz))

    def _ln_total_metabolite_weight(self, ln_conc: cp.Variable) -> Expression:
        """Calculate the enzyme cost for an input concentration profile."""
        return cp.log_sum_exp(ln_conc + np.log(self.mw_met))

    def _create_ecm_problem(
        self, ln_conc: cp.Variable
    ) -> [Objective, Constraint]:
        obj_enz_wgt = cp.Minimize(self._ln_total_enzyme_weight(ln_conc))

        if self.params["objective"] == "enzyme + metabolite":
            obj_met_wgt = cp.Minimize(self._ln_total_metabolite_weight(ln_conc))
            obj_ec = cp.transforms.scalarize.log_sum_exp(
                [obj_enz_wgt, obj_met_wgt], weights=(1.0, 1.0)
            )
        elif self.params["objective"] == "enzyme":
            obj_ec = obj_enz_wgt
        else:
            raise ValueError(
                f"unsupported 'objective' param: {self.params['objective']}"
            )

        constraints = [
            ln_conc >= self.ln_conc_lb,
            ln_conc <= self.ln_conc_ub,
            self._driving_forces(ln_conc) >= self.DRIVING_FORCE_LB,
        ]

        return obj_ec, constraints

    def optimize_ecm(self) -> Tuple[float, np.ndarray]:
        """Minimize enzyme cost.

        Use convex optimization to find the y with the minimal total
        enzyme cost per flux, i.e. sum(ECF(ln_conc)).
        """
        ln_conc = cp.Variable(self.Nc)
        obj_ec, constraints = self._create_ecm_problem(ln_conc)
        prob = cp.Problem(obj_ec, constraints)
        prob.solve()
        if prob.status == cp.OPTIMAL:
            return np.exp(prob.value), ln_conc.value
        elif prob.status == cp.OPTIMAL_INACCURATE:
            logging.warning("ECM solution is 'optimal_inaccurate'")
            return np.exp(prob.value), ln_conc.value
        else:
            raise Exception(prob.status)

    def _calc_z_scores(self, ln_conc: cp.Variable) -> Expression:
        """Calculate individual zscores."""
        return (ln_conc - self.ln_conc_mu) / (self.ln_conc_sigma + 1e-9)

    def _metabolic_adjustment(self, ln_conc: np.ndarray) -> float:
        """Calculate metabolic adjustment score.

        Essentially, it is the sum of Z-scores for the multivariate Gaussian
        distribution of log-concentrations (described by mu and sigma)
        """
        idx = self.ln_conc_sigma >= self.MINIMAL_STDEV
        return sum(cp.power(self._calc_z_scores(ln_conc)[idx], 2))

    def pareto(self, weights: Iterable[float] = None) -> pd.DataFrame:
        """Minimize enzyme cost versus metabolic adjustment (Pareto).

        enzyme cost is defined as in standard ECM.
        metabolic adjustment is the sum of squared Z-scores of the metabolite
        log-concentrations (relative to the prior Gaussian distribution).

        Arguments
        ---------
        weights : Iterable[float]
            the factor for the upper bounds on the ECM objective. Only
            non-negative values are allowed

        Returns
        -------
        results : DataFrame
            a summary table of the results of all optimization runs
        """
        ln_conc = cp.Variable(self.Nc)
        ln_enz_wgt = self._ln_total_enzyme_weight(ln_conc)
        ln_met_wgt = self._ln_total_metabolite_weight(ln_conc)
        constraints = [
            ln_conc >= self.ln_conc_lb,
            ln_conc <= self.ln_conc_ub,
            self._driving_forces(ln_conc) >= self.DRIVING_FORCE_LB,
        ]

        # the two Pareto objectives:
        ln_enz_cst = cp.log_sum_exp(cp.vstack([ln_enz_wgt, ln_met_wgt]))
        met_adj = self._metabolic_adjustment(ln_conc)

        # First find the minimum EC value
        prob = cp.Problem(cp.Minimize(ln_enz_cst), constraints)
        prob.solve()
        ln_min_ec = ln_enz_cst.value

        # scan the Pareto front
        data = []
        if weights is None:
            weights = np.linspace(0, 1, num=30)

        for w in weights:
            # place an upper bound on the ECM objective:
            # ECM < (min(ECM) * exp(w))
            # and the minimize the secondary objective (metabolic adjustment)
            prob = cp.Problem(
                cp.Minimize(met_adj),
                constraints + [ln_enz_cst <= ln_min_ec + w],
            )
            prob.solve()

            if prob.status not in [cp.OPTIMAL, cp.OPTIMAL_INACCURATE]:
                raise Exception(prob.status)

            data += [
                (w, "obj", "enzyme_cost", None, np.exp(ln_enz_cst.value)),
                (w, "obj", "enzyme_weight", None, np.exp(ln_enz_wgt.value)),
                (w, "obj", "metabolite_weight", None, np.exp(ln_met_wgt.value)),
                (w, "obj", "metabolic_adjustment", None, met_adj.value),
            ]

            for j, x in enumerate(ln_conc.value):
                data.append((w, "primal", "log_conc", j, x))
            for j, z_score in enumerate(self._calc_z_scores(ln_conc).value):
                data.append((w, "z_score", "log_conc", j, z_score))
            for i, df in enumerate(self._driving_forces(ln_conc).value):
                data.append((w, "primal", "driving_force", i, df))
            for i, enzyme_conc in enumerate(self.ECF(ln_conc).value):
                data.append((w, "primal", "log_enzyme_conc", i, enzyme_conc))
            for i, eta in enumerate(self._ln_eta_thermodynamic(ln_conc).value):
                data.append((w, "eta", "thermodynamic", i, np.exp(eta)))
            for i, eta in enumerate(self._ln_eta_kinetic(ln_conc).value):
                data.append((w, "eta", "kinetic", i, np.exp(eta)))
            for i, eta in enumerate(self._ln_eta_allosteric(ln_conc).value):
                data.append((w, "eta", "allosteric", i, np.exp(eta)))

        return pd.DataFrame(
            data, columns=["weight", "var_type", "name", "index", "value"]
        )



# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from typing import Tuple

import numpy as np
from scipy.integrate import ode

from .cost_function import EnzymeCostFunction


class EnzymeCostSimulator(object):
    def __init__(self, ecf: EnzymeCostFunction) -> None:
        self.ecf = ecf

        self.idx_fixed = np.where(
            self.ecf.ln_conc_sigma <= self.ecf.MINIMAL_STDEV
        )[0].tolist()

        self.idx_non_fixed = np.where(
            self.ecf.ln_conc_sigma > self.ecf.MINIMAL_STDEV
        )[0].tolist()

        self.E = np.ones(self.ecf.Nr)

    def set_enzyme_concentrations(self, E: np.ndarray) -> None:
        assert E.shape == (self.ecf.Nr,)
        self.E = E

    def _derivaties(self, t, y):
        # we only care about the time derivatives of the internal
        # metabolites (i.e. the first and last one are assumed to be
        # fixed in time)
        ln_conc = np.log(np.array(y))
        v = self.ecf.get_fluxes(ln_conc, self.E)
        dy = self.ecf.S @ v
        dy[self.idx_fixed] = 0
        return dy

    def Simulate(
        self,
        lnC0: np.ndarray,
        t_max: float = 1000.0,
        dt: float = 1.0,
        eps: float = 1e-9,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Find the steady-state solution for the metabolite concentrations
        given the enzyme abundances

        Arguments:
            E    - enzyme abundances [gr]
            y0   - initial concentration of internal metabolites
                   (default: MDF solution)
            eps  - the minimal change under which the simulation will stop

        Returns:
            v    - the steady state flux
            y    - the steady state internal metabolite concentrations
        """
        assert lnC0.shape == (self.ecf.Nc,)

        if not self.ecf.is_feasible(lnC0):
            raise ValueError(
                "initial concentrations are not thermodynamically feasible"
            )

        v = self.ecf.get_fluxes(lnC0, self.E)

        T = np.array([0])
        Y = np.exp(lnC0)
        V = v.T

        r = ode(self._derivaties)
        r.set_initial_value(Y.T, 0)

        while r.successful():
            r.integrate(r.t + dt)
            v = self.ecf.get_fluxes(np.log(r.y.squeeze()), self.E)

            T = np.hstack([T, r.t])
            Y = np.vstack([Y, r.y.T])
            V = np.vstack([V, v.T])

            if r.t >= t_max:
                break

            abs_dy = np.abs(self.ecf.S[self.idx_non_fixed, :] @ v)
            print(f"t = {r.t:.2e}, dy = {abs_dy[0]:.2e}")
            if r.t >= 0.05 * t_max:
                if (abs_dy > eps).any():
                    break

        if r.t >= t_max:
            v_inf = np.nan
            lnC_inf = np.nan
        else:
            v_inf = V[-1, 0]
            lnC_inf = np.log(Y[-1, :])

        return v_inf, lnC_inf



# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import warnings
from typing import Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from equilibrator_api import Q_, ComponentContribution
from sbtab import SBtab

from .cost_function import EnzymeCostFunction
from .ecm_solution import PathwayEcmSolution
from .thermo_models import ThermodynamicModel
from .util import ECF_DEFAULTS


class EnzymeCostModel(object):
    """A class for Enzyme Cost Minimization analysis."""

    DATAFRAME_NAMES = {
        "Compound",
        "Reaction",
        "ConcentrationConstraint",
        "Parameter",
        "Flux",
    }

    def __init__(
        self, thermo_model: ThermodynamicModel, param_df: pd.DataFrame
    ):
        self._thermo_model = thermo_model
        self.config_dict = dict(ECF_DEFAULTS)
        self.config_dict.update(self._thermo_model.config_dict)

        # TODO: use the stdev_factor (and the uncertainty estimates for the
        #  Gibbs energies, to somehow weigh the ECF)
        #  stdev_factor = self.config_dict.get("stdev_factor", 1.0)

        (
            rid2crc_gmean,
            rid2crc_fwd,
            rid2crc_rev,
            rid_cid2KMM,
            rid2mw,
            cid2mw,
        ) = EnzymeCostModel.read_parameters(param_df, self.config_dict)

        self.reaction_ids = list(self._thermo_model.reaction_ids)
        self.compound_ids = list(self._thermo_model.compound_ids)
        assert self._thermo_model.standard_dg_primes.check("kJ/mol")
        standard_dg_primes = np.reshape(
            self._thermo_model.standard_dg_primes, (len(self.reaction_ids), 1)
        )

        KMM = EnzymeCostModel._generate_KMM(
            self.compound_ids, self.reaction_ids, rid_cid2KMM
        )

        # we need all fluxes to be positive, so for every negative flux,
        # we multiply it and the corresponding column in S by (-1)
        dir_mat = np.diag(
            np.sign(self._thermo_model.fluxes.magnitude + 1e-10).flat
        )
        flux = dir_mat @ self._thermo_model.fluxes
        S = self._thermo_model.S.values @ dir_mat
        standard_dg = dir_mat @ standard_dg_primes

        # we only need to define get kcat in the direction of the flux
        # if we use the 'gmean' option, that means we assume we only know
        # the geometric mean of the kcat, and we distribute it between
        # kcat_fwd and kcat_bwd according to the Haldane relationship
        # if we use the 'fwd' option, we just take the kcat in the
        # direction of flux (as is) and that would mean that our
        # thermodynamic rate law would be equivalent to calculating the
        # reverse kcat using the Haldane relationship
        assert self.config_dict["kcat_source"] in ["gmean", "fwd"], (
            "unrecognized kcat source: " + self.config_dict["kcat_source"]
        )

        kcat = []
        for rid, d in zip(
            self._thermo_model.reaction_ids, np.diag(dir_mat).flat
        ):
            if self.config_dict["kcat_source"] == "gmean":
                # TODO: calculate the forward kcat using the Haldane relation
                #  and using the KMM matrix
                crc = rid2crc_gmean[rid]
            elif d >= 0:  # the flux in this reaction is forward
                crc = rid2crc_fwd[rid]
            else:  # the flux in this reaction is backward
                crc = rid2crc_rev[rid]
            kcat.append(crc.m_as("1/s"))
        kcat = Q_(kcat, "1/s")

        # TODO: turn this into a warning, if the MW data is missing, try to
        #  assign a default value
        mw_enz = []
        for rid in self.reaction_ids:
            if rid not in rid2mw:
                raise KeyError(f"This reaction is missing an enzyme MW: {rid}")
            mw_enz.append(rid2mw[rid].m_as("Da"))
        mw_enz = Q_(mw_enz, "Da")

        # TODO: fill gaps in MW data by using the equilibrator-cache
        mw_met = []
        for cid in self.compound_ids:
            if cid not in cid2mw:
                raise KeyError(f"This compound is missing a MW: {cid}")
            mw_met.append(cid2mw[cid].m_as("Da"))
        mw_met = Q_(mw_met, "Da")

        # we must remove H2O from the model, since it should not be considered
        # as a "normal" metabolite in terms of the enzyme and metabolite costs
        self.ecf = EnzymeCostFunction(
            S,
            fluxes=flux,
            kcat=kcat,
            standard_dg=standard_dg,
            KMM=KMM,
            ln_conc_lb=self._thermo_model.ln_conc_lb,
            ln_conc_ub=self._thermo_model.ln_conc_ub,
            mw_enz=mw_enz,
            mw_met=mw_met,
            params=self.config_dict,
            idx_water=self._thermo_model.idx_water,
        )

        self._val_df_dict = None
        self._met_conc_unit = Q_("M")
        self._enz_conc_unit = Q_("M")

    @staticmethod
    def from_sbtab(
        filename: Union[str, SBtab.SBtabDocument],
        comp_contrib: Optional[ComponentContribution] = None,
    ) -> "EnzymeCostModel":
        if isinstance(filename, str):
            sbtabdoc = SBtab.read_csv(filename, "pathway")
        elif isinstance(filename, SBtab.SBtabDocument):
            sbtabdoc = filename

        thermo_model = ThermodynamicModel.from_sbtab(sbtabdoc, comp_contrib)

        param_sbtab = sbtabdoc.get_sbtab_by_id("Parameter")
        assert param_sbtab, "Missing table 'Parameter' in the SBtab document"
        param_df = param_sbtab.to_data_frame()

        return EnzymeCostModel(thermo_model, param_df)

    def add_validation_data(
        self, filename: Union[str, SBtab.SBtabDocument]
    ) -> None:
        if isinstance(filename, str):
            sbtabdoc = SBtab.read_csv(filename, "pathway")
        elif isinstance(filename, SBtab.SBtabDocument):
            sbtabdoc = filename

        conc_sbtab = sbtabdoc.get_sbtab_by_id("Concentration")
        self._met_conc_unit = Q_(conc_sbtab.get_attribute("Unit"))
        assert self._met_conc_unit.check("[concentration]"), (
            "Metabolite concentration unit is not a [concentration] quantity",
            self._met_conc_unit,
        )

        enzyme_sbtab = sbtabdoc.get_sbtab_by_id("EnzymeConcentration")
        self._enz_conc_unit = Q_(enzyme_sbtab.get_attribute("Unit"))
        assert self._enz_conc_unit.check("[concentration]"), (
            "Enzyme concentration unit is not a [concentration] quantity",
            self._enz_conc_unit,
        )

        self._val_df_dict = {
            sbtab.table_id: sbtab.to_data_frame() for sbtab in sbtabdoc.sbtabs
        }

    @staticmethod
    def read_parameters(
        parameter_df: pd.DataFrame, ecf_params: dict
    ) -> Tuple[dict, ...]:
        cols = ["QuantityType", "Value", "Compound", "Reaction", "Unit"]

        rid2mw = dict()
        cid2mw = dict()
        rid2crc_gmean = dict()  # catalytic rate constant geomertic mean
        rid2crc_fwd = dict()  # catalytic rate constant forward
        rid2crc_rev = dict()  # catalytic rate constant reverse
        crctype2dict = {
            "catalytic rate constant geometric mean": rid2crc_gmean,
            "substrate catalytic rate constant": rid2crc_fwd,
            "product catalytic rate constant": rid2crc_rev,
        }

        rid_cid2KMM = {}  # Michaelis-Menten constants

        for i, row in parameter_df.iterrows():
            try:
                typ, val, cid, rid, unit = [row[c] for c in cols]
                val = Q_(float(val), unit)

                if typ in crctype2dict:
                    assert val.check(
                        "1/[time]"
                    ), "rate constants must have inverse time units"
                    val.ito("1/s")
                    crctype2dict[typ][rid] = val
                elif typ == "Michaelis constant":
                    assert val.check(
                        "[concentration]"
                    ), "Michaelis constants must have concentration units"
                    val.ito("molar")
                    rid_cid2KMM[rid, cid] = val
                elif typ == "protein molecular mass":
                    assert val.check("[mass]"), f"'{typ}' must have mass units"
                    val.ito("Da")
                    rid2mw[rid] = val
                elif typ == "molecular mass":
                    assert val.check("[mass]"), f"'{typ}' must have mass units"
                    val.ito("Da")
                    cid2mw[cid] = val
                else:
                    warnings.warn("unrecognized Parameter: " + typ)
            except AssertionError:
                raise ValueError(
                    "Syntax error in Parameter table, row %d - %s" % (i, row)
                )
        # make sure not to count water as contributing to the volume or
        # cost of a reaction
        return (
            rid2crc_gmean,
            rid2crc_fwd,
            rid2crc_rev,
            rid_cid2KMM,
            rid2mw,
            cid2mw,
        )

    @staticmethod
    def _generate_KMM(
        cids: List[str], rids: List[str], rid_cid2KMM: dict
    ) -> np.array:
        KMM = np.ones((len(cids), len(rids)))
        for i, cid in enumerate(cids):
            for j, rid in enumerate(rids):
                kmm = rid_cid2KMM.get((rid, cid), Q_(1, "M"))
                KMM[i, j] = kmm.m_as("M")
        return KMM * Q_(1, "M")

    def optimize_ecm(self) -> PathwayEcmSolution:
        return PathwayEcmSolution(self, *self.ecf.optimize_ecm())

    def pareto(self, weights: Iterable[float] = None) -> pd.DataFrame:
        return self.ecf.pareto(weights=weights)

    def _get_measured_metabolite_conc(self) -> Q_:
        assert (
            self._val_df_dict is not None
        ), "cannot validate results because no validation data was given"

        met_conc_df = self._val_df_dict["Concentration"].set_index("Compound")
        met_concentrations = Q_(
            pd.to_numeric(met_conc_df.Value)[self.compound_ids].values,
            self._met_conc_unit,
        )
        return met_concentrations

    def _get_measured_enzyme_conc(self) -> Q_:
        assert (
            self._val_df_dict is not None
        ), "cannot validate results because no validation data was given"

        enz_conc_df = self._val_df_dict["EnzymeConcentration"].set_index(
            "Reaction"
        )
        enz_concentrations = Q_(
            pd.to_numeric(enz_conc_df.Value)[self.reaction_ids].values,
            self._met_conc_unit,
        )
        return enz_concentrations



"""A class for visualizing the MDMC solution."""
# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from typing import Optional

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt


class PathwayMdmcSolution(object):
    """Handle MDMC results.

    PathwayMDFData is a container class for MDF results, with plotting
    capabilities.
    """

    def __init__(self, thermo_model, solution_df: pd.DataFrame) -> None:
        """Create PathwayMDFData object.

        :param thermo_model: a PathwayThermoModel object
        :param df: a DataFrame with all the QP results (for all lower bounds)
        """
        self.reaction_ids = list(thermo_model.reaction_ids)
        self.compound_ids = list(thermo_model.compound_ids)

        self.solution_df = solution_df
        self.solution_df["df_lb"] = self.solution_df.df_lb.round(2)

        self.min_lb = self.solution_df.df_lb.min()
        self.max_lb = self.solution_df.df_lb.max()

        # replace the internal QP names with reaction IDs and compound names
        self.solution_df["name"] = ""

        var_index_to_cpd_name = dict(enumerate(self.compound_ids))
        var_idx = self.solution_df.var_type == "var"
        self.solution_df.loc[var_idx, "name"] = self.solution_df.loc[
            var_idx, "index"
        ].apply(var_index_to_cpd_name.get)

        cnstr_index_to_rxn_name = dict(enumerate(self.reaction_ids))
        cnstr_idx = self.solution_df.var_type == "cnstr"
        self.solution_df.loc[cnstr_idx, "name"] = self.solution_df.loc[
            cnstr_idx, "index"
        ].apply(cnstr_index_to_rxn_name.get)

        self.objective_values = (
            self.solution_df[self.solution_df.var_type == "obj"]
            .set_index("df_lb")
            .value
        )
        self.objective_values.name = "Total squared Z-scores"

        grouped_df = self.solution_df.groupby(
            ["value_type", "var_type", "var_name"]
        )

        self.concentration_df = grouped_df.get_group(
            ("primal", "var", "log_conc")
        ).pivot("name", "df_lb", "value")
        self.concentration_df = np.exp(self.concentration_df)

        self.zscore_df = grouped_df.get_group(
            ("zscore", "var", "log_conc")
        ).pivot("name", "df_lb", "value")

        self._reaction_prices_df = grouped_df.get_group(
            ("shadow_price", "cnstr", "driving_force")
        ).pivot("name", "df_lb", "value")

    def reaction_prices_df(self, normalized=False) -> pd.DataFrame:
        """Get the reaction shadow prices."""
        abs_prices = self._reaction_prices_df.apply(np.abs)
        if normalized:
            # normalize the shadow prices in each column to 1
            return abs_prices.multiply(1.0 / abs_prices.sum(0))
        else:
            return abs_prices

    def plot_objectives(self, ax: Optional[plt.Axes] = None) -> None:
        """Plot the Pareto front of the two objectives."""
        if not ax:
            fig, ax = plt.subplots(1, 1, figsize=(6, 6))

        self.objective_values.plot(style="b-", ax=ax, legend=False)
        ax.set_xlabel("lower bound on all Driving Forces [kJ/mol]")
        ax.set_ylabel("Sum of squared Z-scores")

    def plot_reaction_prices(
        self, ax: Optional[plt.Axes] = None, normalized: bool = False
    ) -> None:
        """Plot compound concentrations."""
        if not ax:
            fig, ax = plt.subplots(1, 1, figsize=(8, 8))
        _df = self.reaction_prices_df(normalized=normalized)

        _df.T.plot(ax=ax, legend=False)
        for label, yvalue in _df.iloc[:, 0].items():
            ax.text(self.min_lb - 0.1, yvalue, label, ha="right", va="center")
        for label, yvalue in _df.iloc[:, -1].items():
            ax.text(self.max_lb + 0.1, yvalue, label, ha="left", va="center")
        if normalized:
            ax.set_ylabel("normalized shadow price")
        else:
            ax.set_ylabel("shadow price")
        ax.set_xlabel("lower bound on all Driving Forces [kJ/mol]")
        ax.set_xlim(self.min_lb - 1, self.max_lb + 1)

    def plot_concentrations(
        self, ax: Optional[plt.Axes] = None, show_zscores: bool = True
    ) -> None:
        """Plot all compound concentrations (and, optionally, Z-scores)."""
        if not ax:
            fig, ax = plt.subplots(1, 1, figsize=(15, 8))
        if show_zscores:
            ax.plot(
                self.concentration_df.columns,
                self.concentration_df.values.T,
                color=(0.6, 0.6, 0.6),
                zorder=1,
            )
        else:
            self.concentration_df.T.plot(ax=ax, legend=False)

        for met, row in self.concentration_df.iterrows():
            if show_zscores:
                zscores = self.zscore_df.loc[met, :]
                sc = ax.scatter(
                    row.index,
                    row.values,
                    c=zscores,
                    s=5,
                    cmap="coolwarm",
                    vmin=-4,
                    vmax=4,
                    zorder=2,
                )
            ax.text(self.min_lb - 0.1, row.iat[0], met, ha="right", va="center")
            ax.text(self.max_lb + 0.1, row.iat[-1], met, ha="left", va="center")
        ax.set_ylabel("concentration [M]")
        ax.set_xlabel("lower bound on all Driving Forces [kJ/mol]")
        ax.set_yscale("log")
        ax.set_xlim(self.min_lb - 2, self.max_lb + 2)
        if show_zscores:
            plt.colorbar(sc, label="Z-score")

    def plot_metabolite(self, met: str) -> plt.Figure:
        """Plot a single compound's concentration."""
        from mpl_toolkits.mplot3d import Axes3D

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        # ax = fig.gca(projection='3d')

        concentrations = self.concentration_df.loc[met, :]
        zscores = self.zscore_df.loc[met, :]
        sc = ax.scatter(
            concentrations.index,
            concentrations.values,
            c=zscores.values,
            linewidth=0,
            s=5,
        )
        # ax.text(self.min_lb - 0.1, concentrations.iat[0], met, ha="right", va="center")
        # ax.text(self.max_lb + 0.1, concentrations.iat[-1], met, ha="left", va="center")
        ax.set_ylabel("{met} concentration [M]")
        ax.set_xlabel("lower bound on all Driving Forces [kJ/mol]")
        ax.set_yscale("log")
        ax.set_xlim(self.min_lb - 2, self.max_lb + 2)
        return fig
